#include "IZ_joystick.h"

void IZ_HandleJoystickDeviceEvents(SDL_Event e, IZ_JoystickState* state) {
	if (e.type == SDL_JOYDEVICEADDED) {
		if (SDL_NumJoysticks() <= PLAYERS && !state->device) {
			state->device = SDL_JoystickOpen(e.jdevice.which);
		}
		return;
	}

	if (e.type == SDL_JOYDEVICEREMOVED) {
		if (
			state->device
			&& SDL_JoystickInstanceID(state->device) == e.jdevice.which
		) {
			state->device = NULL;
		}
	}
}

void IZ_HandleJoystickAxisEvents(SDL_Event e, IZ_JoystickState* state, IZ_Action* action) {
	if (e.type == SDL_JOYAXISMOTION) {
		if (
			e.jaxis.axis == IZ_JOY_AXIS_DIRECTION_HORIZONTAL1
			|| e.jaxis.axis == IZ_JOY_AXIS_DIRECTION_HORIZONTAL2
		) {
			*action &= ~(0x1 << IZ_ACTION_INDEX_RIGHT);
			*action &= ~(0x1 << IZ_ACTION_INDEX_LEFT);
			if (e.jaxis.value > state->config.axis_threshold) {
				*action |= (0x1 << IZ_ACTION_INDEX_RIGHT);
				return;
			}
			if (e.jaxis.value <= -state->config.axis_threshold) {
				*action |= (0x1 << IZ_ACTION_INDEX_LEFT);
			}
			return;
		}
		if (
			e.jaxis.axis == IZ_JOY_AXIS_DIRECTION_VERTICAL1
			|| e.jaxis.axis == IZ_JOY_AXIS_DIRECTION_VERTICAL2
		) {
			*action &= ~(0x1 << IZ_ACTION_INDEX_UP);
			*action &= ~(0x1 << IZ_ACTION_INDEX_DOWN);
			if (e.jaxis.value > state->config.axis_threshold) {
				*action |= (0x1 << IZ_ACTION_INDEX_DOWN);
				return;
			}
			if (e.jaxis.value <= -state->config.axis_threshold) {
				*action |= (0x1 << IZ_ACTION_INDEX_UP);
			}
		}
	}
}

void IZ_HandleJoystickHatEvents(SDL_Event e, IZ_Action* action) {
	if (e.type == SDL_JOYHATMOTION) {
		*action &= ~(0x1 << IZ_ACTION_INDEX_UP);
		*action &= ~(0x1 << IZ_ACTION_INDEX_RIGHT);
		*action &= ~(0x1 << IZ_ACTION_INDEX_DOWN);
		*action &= ~(0x1 << IZ_ACTION_INDEX_LEFT);
		if (e.jhat.value != 0) {
			*action |= e.jhat.value;
		}
	}
}

void IZ_HandleJoystickButtonEvents(SDL_Event e, IZ_JoystickState* state, IZ_Action* action) {
	for (uint8_t i = 4; i < CONTROLS; i += 1) {
		if (e.jbutton.button == state->config.control_mapping[i]) {
			const uint16_t bitflag = (0x1 << i);

			if (e.type == SDL_JOYBUTTONDOWN) {
				*action |= bitflag;
				return;
			}

			if (e.type == SDL_JOYBUTTONUP) {
				*action &= ~bitflag;
				return;
			}
		}
	}
}

void IZ_HandleJoystickEvents(SDL_Event e, IZ_JoystickState(* state)[PLAYERS], IZ_Action(* action)[PLAYERS]) {
	for (uint8_t player_index = 0; player_index < PLAYERS; player_index += 1) {
		IZ_HandleJoystickDeviceEvents(e, state[player_index]);
		IZ_HandleJoystickAxisEvents(e, state[player_index], action[player_index]);
		IZ_HandleJoystickHatEvents(e, action[player_index]);
		IZ_HandleJoystickButtonEvents(e, state[player_index], action[player_index]);
	}
}

void IZ_LoadJoystickConfig(const char* config_path, IZ_JoystickState(* state)[PLAYERS]) {
	char control_mapping_section_name[26];
	char main_section_name[11];
	for (uint8_t player_index = 0; player_index < PLAYERS; player_index += 1) {
		sprintf_s(control_mapping_section_name, 26, "Joystick.%d.ControlMapping", player_index);

		for (uint8_t i = 4; i < CONTROLS; i += 1) {
			state[player_index]->config.control_mapping[i] = ini_getl(
				control_mapping_section_name,
				ACTION_NAMES[i],
				IZ_DEFAULT_JOYSTICK_CONTROLS[player_index][i],
				config_path
			);
		}

		sprintf_s(main_section_name, 11, "Joystick.%d", player_index);
		state[player_index]->config.axis_threshold = ini_getl(main_section_name, "AxisThreshold", IZ_DEFAULT_AXIS_THRESHOLD, config_path);
		state[player_index]->config.device_id = ini_getl(main_section_name, "DeviceID", player_index, config_path);
	}
}

IZ_ProcedureResult IZ_SaveJoystickConfig(const char* config_path, IZ_JoystickState(* state)[PLAYERS]) {
	uint8_t problem = 0;

	char control_mapping_section_name[26];
	char main_section_name[11];

	for (uint8_t player_index = 0; player_index < PLAYERS; player_index += 1) {
		sprintf_s(control_mapping_section_name, 26, "Joystick.%d.ControlMapping", player_index);
		for (uint8_t i = 4; i < CONTROLS; i += 1) {
			if (!ini_putl(
				control_mapping_section_name,
				ACTION_NAMES[i],
				state[player_index]->config.control_mapping[i],
				config_path
			)) {
				return 1;
			}
		}

		sprintf_s(main_section_name, 11, "Joystick.%d", player_index);
		if (!ini_putl(
			main_section_name,
			"DeviceID",
			state[player_index]->config.device_id,
			config_path
		)) {
			problem |= (1 << player_index);
		}

		if (!ini_putl(
			main_section_name,
			"AxisThreshold",
			state[player_index]->config.axis_threshold,
			config_path
		)) {
			problem |= (1 << player_index);
		}
	}

	return problem;
}

IZ_ProcedureResult IZ_InitializeJoystickState(const char* config_path, IZ_JoystickState(* state)[PLAYERS]) {
	IZ_LoadJoystickConfig(config_path, state);
	if (IZ_SaveJoystickConfig(config_path, state)) {
		return 1;
	}

	uint8_t joysticks_count = SDL_NumJoysticks();
	for (uint8_t player_index = 0; player_index < joysticks_count; player_index += 1) {
		state[player_index]->device = SDL_JoystickOpen(state[player_index]->config.device_id);
	}

	return 0;
}

void IZ_TeardownJoystickState(IZ_JoystickState(* state)[PLAYERS]) {
	for (uint8_t player_index = 0; player_index < PLAYERS; player_index += 1) {
		if (!state[player_index]->device) {
			continue;
		}
		SDL_JoystickClose(state[player_index]->device);
	}
}
